﻿unit uFOV;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Math;

type
  TSetXYVal = procedure(X, Y: integer; Value: boolean) of object;
  TGetXYVal = function(X, Y: integer): boolean of object;

procedure InitRPAS(XSize, YSize: integer);
procedure RPAS(GetTrans, GetFov: TGetXYVal; SetFov: TSetXYVal;
  PlayerX, PlayerY, MaxRadius: integer; LightWalls: boolean);

function CRound(X: single): integer;

implementation

var
  StartAngle, EndAngle:   array[0..(100 * 100) div 7] of single;
  RPAS_MAX_X, RPAS_MAX_Y: integer;

procedure InitRPAS(XSize, YSize: integer);
begin
  RPAS_MAX_X := XSize;
  RPAS_MAX_Y := YSize;
end;

procedure ComputeQuadrant(GetTrans, GetFov: TGetXYVal; SetFov: TSetXYVal;
  PlayerX, PlayerY, MaxRadius: integer; LightWalls: boolean; dx, dy: integer);
var
  Iteration: integer;
  Done: boolean;
  TotalObstracles: integer;
  ObstraclesInLastLine: integer;
  MinAngle: single;
  X, Y: integer;
  SlopesPerCell: single;
  HalfSlopes: single;
  ProcessedCell: integer;
  MinX, MaxX: integer;
  Visible: boolean;
  StartSlope, CenterSlope, EndSlope: single;
  idx: integer;
  MinY, MaxY: integer;
begin
  { octant: vertical edge: }
  Iteration := 1;
  Done := False;
  TotalObstracles := 0;
  ObstraclesInLastLine := 0;
  MinAngle := 0.0;
  { do while there are unblocked slopes left and the algo is within the map's boundaries
    scan progressive lines/columns from the PC outwards
   }
  y := PlayerY + dy;
  if not InRange(y, 1, RPAS_MAX_Y) then
    done := True;
  while not Done do
  begin
    { process cells in the line }
    SlopesPerCell := 1 / (Iteration + 1);
    HalfSlopes := SlopesPerCell * 0.5;
    ProcessedCell := CRound(MinAngle / SlopesPerCell);
    MinX := MAX(1, PlayerX - Iteration);
    MaxX := MIN(RPAS_MAX_X, PlayerX + iteration);
    done := True;
    X    := PlayerX + ProcessedCell * dx;
    while InRange(X, MinX, MaxX) do
    begin
      { calculate slopes per cell }
      Visible     := True;
      StartSlope  := ProcessedCell * SlopesPerCell;
      CenterSlope := StartSlope + HalfSlopes;
      EndSlope    := StartSlope + SlopesPerCell;
      if (ObstraclesInLastLine > 0) and not GetFov(X, Y) then
      begin
        idx := 0;
        while Visible and (idx < ObstraclesInLastLine) do
        begin
          if GetTrans(X, Y) then
          begin
            if (CenterSlope > StartAngle[idx]) and (CenterSlope < EndAngle[idx]) then
              Visible := False;
          end
          else
          begin
            if (StartSlope >= StartAngle[idx]) and (EndSlope <= EndAngle[idx]) then
              Visible := False;
          end;
          if Visible and (not GetFov(X, Y - DY) or not GetTrans(X, Y - DY)) and
            (InRange(x - dx, 1, RPAS_MAX_X) and
            (not GetFov(X - dx, Y - DY) or not GetTrans(X - dx, Y - DY))) then
            Visible := False;
          Inc(idx);
        end;
      end;
      if Visible then
      begin
        SetFov(X, Y, True);
        Done := False;
        { if the cell is opaque, block the adjacent slopes}
        if not GetTrans(X, Y) then
        begin
          if MinAngle >= StartSlope then
            MinAngle := EndSlope
          else
          begin
            StartAngle[TotalObstracles] := StartSlope;
            EndAngle[TotalObstracles]   := EndSlope;
            Inc(TotalObstracles);
          end;
          if not LightWalls then
            SetFov(X, Y, False);
        end;
      end;
      Inc(X, dx);
      Inc(ProcessedCell);
    end;
    if Iteration = MaxRadius then
      Done := True;
    Inc(iteration);
    ObstraclesInLastLine := TotalObstracles;
    Inc(y, dy);
    if not InRange(y, 1, RPAS_MAX_Y) then
      Done := True;
    if MinAngle = 1.0 then
      Done := True;
  end;

  { octant: horizontal edge }
  Iteration := 1; // iteration of the algo for this octant
  Done := False;
  TotalObstracles := 0;
  ObstraclesInLastLine := 0;
  MinAngle := 0;
    { do while there are unblocked slopes left and the algo is within the map's boundaries
      scan progressive lines/columns from the PC outwards
     }
  x := PlayerX + dx; // the outer slope's coordinates (first processed line)
  if not InRange(X, 1, RPAS_MAX_X) then
    Done := True;
  while not Done do
  begin
    { process cells in the line }
    SlopesPerCell := 1 / (Iteration + 1);
    HalfSlopes := SlopesPerCell * 0.5;
    ProcessedCell := CRound(MinAngle / SlopesPerCell);
    MinY := MAX(1, PlayerY - Iteration);
    MaxY := MIN(RPAS_MAX_Y, PlayerY + iteration);
    done := True;
    Y    := PlayerY + ProcessedCell * dy;
    while InRange(y, MinY, MaxY) do
    begin
      { calculate slopes per cell }
      Visible     := True;
      StartSlope  := ProcessedCell * SlopesPerCell;
      CenterSlope := StartSlope + HalfSlopes;
      EndSlope    := StartSlope + SlopesPerCell;
      if (ObstraclesInLastLine > 0) and not GetFov(X, Y) then
      begin
        idx := 0;
        while Visible and (idx < ObstraclesInLastLine) do
        begin
          if GetTrans(X, Y) then
          begin
            if (CenterSlope > StartAngle[idx]) and (CenterSlope < EndAngle[idx]) then
              Visible := False;
          end
          else
          begin
            if (StartSlope >= StartAngle[idx]) and (EndSlope <= EndAngle[idx]) then
              Visible := False;
          end;
          if Visible and (not GetFov(X - dx, Y) or not
            GetTrans(X - dx, Y)) and (InRange(Y - dy, 1, RPAS_MAX_Y) and
            (not GetFov(X - dx, Y - DY) or not GetTrans(X - dx, Y - DY))) then
            Visible := False;
          Inc(idx);
        end;
      end;
      if Visible then
      begin
        SetFov(X, Y, True);
        Done := False;
        { if the cell is opaque, block the adjacent slopes }
        if not GetTrans(X, Y) then
        begin
          if MinAngle >= StartSlope then
            MinAngle := EndSlope
          else
          begin
            StartAngle[TotalObstracles] := StartSlope;
            EndAngle[TotalObstracles]   := EndSlope;
            Inc(TotalObstracles);
          end;
          if not LightWalls then
            SetFov(X, Y, False);
        end;
      end;
      Inc(ProcessedCell);
      Inc(Y, dy);
    end;
    if Iteration = MaxRadius then
      Done := True;
    Inc(iteration);
    ObstraclesInLastLine := TotalObstracles;
    Inc(x, dx);
    if not InRange(x, 1, RPAS_MAX_X) then
      done := True;
    if MinAngle = 1.0 then
      Done := True;
  end;
end;

procedure RPAS(GetTrans, GetFov: TGetXYVal; SetFov: TSetXYVal;
  PlayerX, PlayerY, MaxRadius: integer; LightWalls: boolean);
var
  I, J: integer;
begin
  for I := 1 to RPAS_MAX_X do
    for J := 1 to RPAS_MAX_Y do
      SetFov(I, J, False);
  SetFov(PlayerX, PlayerY, True);
  ComputeQuadrant(GetTrans, GetFov, SetFov, PlayerX, PlayerY, MaxRadius,
    LightWalls, 1, 1);
  ComputeQuadrant(GetTrans, GetFov, SetFov, PlayerX, PlayerY, MaxRadius,
    LightWalls, 1, -1);
  ComputeQuadrant(GetTrans, GetFov, SetFov, PlayerX, PlayerY, MaxRadius,
    LightWalls, -1, 1);
  ComputeQuadrant(GetTrans, GetFov, SetFov, PlayerX, PlayerY, MaxRadius,
    LightWalls, -1, -1);
  //  if MaxRadius > 0 then
  for I := 1 to RPAS_MAX_X do
    for J := 1 to RPAS_MAX_Y do
      if Distance(PlayerX, PlayerY, I, J) > MaxRadius then
        SetFov(I, J, False);
end;

function CRound(X: single): integer;
begin
  Result := Floor(X);
end;

end.
